import torchvision.transforms as T
from torchvision import datasets

from ..utils import common_functions as c_f
from ..utils.constants import IMAGENET_MEAN, IMAGENET_STD
from ..utils.transforms import GrayscaleToRGB
from .combined_source_and_target import CombinedSourceAndTargetDataset
from .concat_dataset import ConcatDataset
from .mnistm import MNISTM
from .office31 import Office31
from .officehome import OfficeHome
from .source_dataset import SourceDataset
from .target_dataset import TargetDataset


def get_multiple(dataset_getter, domains, *args):
    return ConcatDataset([dataset_getter(d, *args) for d in domains])


def get_datasets(
    dataset_getter,
    src_domains,
    target_domains,
    folder,
    download=False,
    return_target_with_labels=False,
    transform_getter=None,
):
    def getter(domains, train, is_training):
        return get_multiple(
            dataset_getter,
            domains,
            train,
            is_training,
            folder,
            download,
            transform_getter,
        )

    if not src_domains and not target_domains:
        raise ValueError(
            "At least one of src_domains and target_domains must be provided"
        )

    output = {}
    if src_domains:
        output["src_train"] = SourceDataset(getter(src_domains, True, False))
        output["src_val"] = SourceDataset(getter(src_domains, False, False))
    if target_domains:
        output["target_train"] = TargetDataset(getter(target_domains, True, False))
        output["target_val"] = TargetDataset(getter(target_domains, False, False))
        if return_target_with_labels:
            output["target_train_with_labels"] = SourceDataset(
                getter(target_domains, True, False), domain=1
            )
            output["target_val_with_labels"] = SourceDataset(
                getter(target_domains, False, False), domain=1
            )
    if src_domains and target_domains:
        output["train"] = CombinedSourceAndTargetDataset(
            SourceDataset(getter(src_domains, True, True)),
            TargetDataset(getter(target_domains, True, True)),
        )
    elif src_domains:
        output["train"] = SourceDataset(getter(src_domains, True, True))
    elif target_domains:
        output["train"] = TargetDataset(getter(target_domains, True, True))
    return output


def get_mnist_transform(domain, *_):
    if domain == "mnist":
        return T.Compose(
            [
                T.Resize(32),
                T.ToTensor(),
                GrayscaleToRGB(),
                T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
            ]
        )
    elif domain == "mnistm":
        return T.Compose(
            [
                T.ToTensor(),
                T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
            ]
        )


def _get_mnist_mnistm(domain, train, is_training, folder, download, transform_getter):
    transform_getter = c_f.default(transform_getter, get_mnist_transform)
    transform = transform_getter(domain, train, is_training)
    if domain == "mnist":
        return datasets.MNIST(
            folder, train=train, transform=transform, download=download
        )
    elif domain == "mnistm":
        return MNISTM(folder, train, transform, download=download)


def get_mnist_mnistm(*args, **kwargs):
    return get_datasets(_get_mnist_mnistm, *args, **kwargs)


def get_resnet_transform(domain, train, is_training):
    transform = [T.Resize(256)]
    if is_training:
        transform += [
            T.RandomCrop(224),
            T.RandomHorizontalFlip(),
        ]
    else:
        transform += [T.CenterCrop(224)]

    transform += [
        T.ToTensor(),
        T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
    ]
    return T.Compose(transform)


def standard_dataset(cls):
    def fn(domain, train, is_training, folder, download, transform_getter):
        transform_getter = c_f.default(transform_getter, get_resnet_transform)
        transform = transform_getter(domain, train, is_training)
        return cls(
            root=folder,
            domain=domain,
            train=train,
            transform=transform,
            download=download,
        )

    return fn


def get_office31(*args, **kwargs):
    return get_datasets(standard_dataset(Office31), *args, **kwargs)


def get_officehome(*args, **kwargs):
    return get_datasets(standard_dataset(OfficeHome), *args, **kwargs)



from .classifier import Classifier
from .mnist import MNISTFeatures
from .utils import download_weights


def mnistG(pretrained=False, progress=True):
    """
    Returns:
        An [```MNISTFeatures```][pytorch_adapt.models.MNISTFeatures] model
        trained on the MNIST dataset, if ```pretrained == True```.
    """
    model = MNISTFeatures()
    url = "https://cornell.box.com/shared/static/tdx0ts24e273j7mf3r2ox7a12xh4fdfy"
    h = "68ee79452f1d5301be2329dfa542ac6fa18de99e09d6540838606d9d700b09c8"
    filename = f"mnistG-{h[:8]}.pth"
    return download_weights(model, url, pretrained, progress, filename)


def mnistC(num_classes=10, in_size=1200, h=256, pretrained=False, progress=True):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
        trained on the MNIST dataset, if ```pretrained == True```.
    """
    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    url = "https://cornell.box.com/shared/static/j4zrogronmievq1csulrkai7zjm27gcq"
    h = "ac7b5a13df2ef3522b6550a147eb44dde8ff4fead3ddedc540d9fe63c9d597c1"
    filename = f"mnistC-{h[:8]}.pth"
    return download_weights(model, url, pretrained, progress, filename)


def resnet50(pretrained=False, progress=True):
    import timm

    model = timm.create_model("resnet50", pretrained=False, num_classes=0)

    # G was frozen during finetuning
    # So the model for all 3 domains is the same
    url = "https://cornell.box.com/shared/static/1oxb5xk5dq3od1d3gprigznxmqb3wgr1"
    h = "a567ecd6ea5addf29ccfa8bf706be78a35adff7cd5cf5a3a99d89b19807454ae"
    filename = f"resnet50MusgraveUDA-{h[:8]}.pth"
    return download_weights(model, url, pretrained, progress, filename)


def office31G(*args, **kwargs):
    """
    Returns:
        A ResNet50 model trained on ImageNet, if ```pretrained == True```.
    """
    return resnet50(*args, **kwargs)


def office31C(
    domain=None, num_classes=31, in_size=2048, h=256, pretrained=False, progress=True
):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
        trained on the specified ```domain``` of the [Office31][pytorch_adapt.datasets.Office31]
        dataset, if ```pretrained == True```. For example

        ```python
        model = office31(domain="amazon", pretrained=True)
        ```
    """
    if pretrained and not domain:
        raise ValueError("if pretrained, domain must be specified")

    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    if not pretrained:
        return model
    url = {
        "amazon": "https://cornell.box.com/shared/static/6h165jqlxcpo16jbs3a7vpvslb6u9vaq",
        "dslr": "https://cornell.box.com/shared/static/t97sedzf4wrto3yfvr8hxivyblqkljiq",
        "webcam": "https://cornell.box.com/shared/static/zuv7be39v8bijwggrvfzlyw1h0pfwrb4",
    }[domain]
    h = {
        "amazon": "6e2fb6f392538172515c2c673a8b3ead7aad8b88b44aad6468c7e9b11761b667",
        "dslr": "fc0acd7a71eb5f12d4af619e5c63bcc42e5a23441bbd105fe0f7a37c26f37d80",
        "webcam": "b2bb55978380fa9ca6452cba30e0ac2a19b7166d8348bcc1554fdabd185e4cdd",
    }[domain]
    filename = f"office31C{domain}-{h[:8]}.pth"
    return download_weights(model, url, pretrained, progress, filename)


def officehomeG(*args, **kwargs):
    """
    Returns:
        A ResNet50 model trained on ImageNet, if ```pretrained == True```.
    """
    return resnet50(*args, **kwargs)


def officehomeC(
    domain=None, num_classes=65, in_size=2048, h=256, pretrained=False, progress=True
):
    """
    Returns:
        A [```Classifier```][pytorch_adapt.models.Classifier] model
        trained on the specified ```domain``` of the [OfficeHome][pytorch_adapt.datasets.OfficeHome]
        dataset, if ```pretrained == True```. For example

        ```python
        model = officehomeC(domain="art", pretrained=True)
        ```
    """
    if pretrained and not domain:
        raise ValueError("if pretrained, domain must be specified")

    model = Classifier(num_classes=num_classes, in_size=in_size, h=h)
    if not pretrained:
        return model
    url = {
        "art": "https://cornell.box.com/shared/static/wxg7v32e2m0jcmq53amhdipty9veb2xx",
        "clipart": "https://cornell.box.com/shared/static/4dhwhj6fkzg9lfgu0mfskt2kby8mznez",
        "product": "https://cornell.box.com/shared/static/r6f3ltgve5g2lrcdtoykj84rlyzqs6ga",
        "real": "https://cornell.box.com/shared/static/1lf1foq65m77pdpc50isdgsc8k71ei29",
    }[domain]
    h = {
        "art": "8db546ff250d2b54899f92e482e80a68411cbe525134d429987b57d3b0571e4b",
        "clipart": "8e145cc6d2df3ff428aeafa43066bbde97d56e9f844b34408bdca74125e62590",
        "product": "472ff36fdf13ec6c1fa1236d1d0800e2a5cf2e3d366b6b63ff5807dff6a761d8",
        "real": "f0c8d6e941d4f488ff2438eb5cccdc59e78f35961e48f03d2186752e5878c697",
    }[domain]
    filename = f"officehomeC{domain}-{h[:8]}.pth"
    return download_weights(model, url, pretrained, progress, filename)
